import argparse

class _ModularHelpEnabler(argparse.Action):
    """A custom action that enables help for a single argument group."""
    def __call__(self, parser, namespace, values, option_string = None):
        parser.enable_modular_help(self.const)

class ModularArgumentParser(argparse.ArgumentParser):
    """A variant of argparse.ArgumentParser that allows to modularly enable and disable printing help for individual argument groups.
    A new variant of add_argument_groups() allows adding commands of the form --help-group.
    """
    def __init__(self, *args, **kwargs):
        """As argparse.ArgumentParser.__init__(), additionally modular_action_groups = [] can be used to specify the names of argument groups used for the custom help options."""
        self._modular_action_groups = {}
        self._modular_active = False
        modular_help_groups = kwargs.pop('modular_help_groups', [])
        help_all = kwargs.pop('help_all', True)
        super().__init__(*args, **kwargs)
        self._modular_help_groups = {}
        for name in modular_help_groups:
            self._modular_help_groups[name] = self.add_argument_group(name)
        if help_all:
            self.add_argument(
                '--help-all',
                action=_ModularHelpEnabler, nargs=0, default=argparse.SUPPRESS,
                const=None,
                help='show all help messages')

    def add_argument_group(self, *args, **kwargs):
        """As argparse.ArgumentParser.add_argument_group(), additionally help_name, help_group and help_text allow to add a new option `--help-{{help_name}}."""
        help_name = kwargs.pop('help_name', None)
        help_group = kwargs.pop('help_group', None)
        help_text = kwargs.pop('help_text', 'show help for {}')
        grp = super().add_argument_group(*args, **kwargs)
        if help_name is not None:
            self._modular_action_groups[grp] = help_name
            parser = self
            if help_group is not None:
                parser = self._modular_help_groups[help_group]
            parser.add_argument('--help-{}'.format(help_name), action=_ModularHelpEnabler, nargs=0,
                                default=argparse.SUPPRESS, const=grp, help = help_text.format(help_name))
        return grp
    def enable_modular_help(self, grp):
        """Remove grp from self._modular_action_groups and thereby enable printing it."""
        self._modular_active = True
        if grp is None:
            self._modular_action_groups = {}
        else:
            del self._modular_action_groups[grp]
    def parse_args(self, *args, **kwargs):
        res = super().parse_args(*args, **kwargs)
        if self._modular_active:
            self.parse_args(['--help'])
        return res
    def error(self, message):
        if not self._modular_active:
            super().error(message)
    def format_help(self):
        """Removes all items from self._action_groups that are still in self._modular_action_groups."""
        self._action_groups = [
            ag for ag in self._action_groups
            if ag not in self._modular_action_groups
        ]
        return super().format_help()



import logging
import re
import sys
import textwrap

from . import options

def lexer(text):
    __tokens = [
        ('COMMENT', ';[^\n]*'),
        ('LPAREN', '\\('),
        ('RPAREN', '\\)'),
        ('STRINGLIT', '"[^"]*"'),
        ('QUOTEDSYM', '\\|[^\\|]*\\|'),
        ('SYMBOL', '[:a-zA-Z0-9~!@$#%\\^&*_+=<>.?/-]+'),
        ('SPACE', '\\s+'),
        ('MISMATCH', '.'),
    ]
    __token_re = re.compile('|'.join(['(?P<{}>{})'.format(tok[0], tok[1]) for tok in __tokens]))

    for m in __token_re.finditer(text):
        kind = m.lastgroup
        if kind in ['SPACE', 'COMMENT']:
            continue
        if kind == 'MISMATCH':
            logging.warning('Unexpected  {}'.format(m.group()))
        yield m.group()

def parse_expression(tokens):
    """Parses a single S-sxpression from a token stream. Uses a simple iterative variant of a recursive descent parser with only a single parsing rule."""
    stack = []
    while True:
        tok = next(tokens)
        if tok == '(':
            stack.append([])
        elif tok == ')':
            cur = stack.pop()
            if not stack:
                return cur
            stack[-1].append(cur)
        else:
            stack[-1].append(tok)
    return None

def parse_smtlib(text):
    """Parses an SMT-LIB input to a sequence of nodes."""
    token_stream = lexer(text)
    exprs = []
    while True:
        try:
            exprs.append(parse_expression(token_stream))
        except StopIteration:
            break
        except RuntimeError as err:
            if isinstance(err.__cause__, StopIteration):
                break
            raise
    return exprs

def render_expression(expr):
    """Renders a node to a string."""
    if isinstance(expr, list):
        return '(' + ' '.join(map(render_expression, expr)) + ')'
    return expr

def render_pretty_expression(expr, indent = ''):
    """Renders a node to a string in a pretty way."""
    if isinstance(expr, list):
        if expr != [] and expr[0] in ['declare-const', 'declare-fun']:
            return render_expression(expr)
        if all(map(lambda e: not isinstance(e, list), expr)):
            return '(' + ' '.join(expr) + ')'
        res = '(' + render_pretty_expression(expr[0]) + '\n'
        for e in expr[1:]:
            res += indent + '\t' + render_pretty_expression(e, indent + '\t') + '\n'
        res += indent + ')'
        return res
    return expr

def render_smtlib(exprs):
    """Renders a sequence of nodes to a string."""
    if options.args().pretty_print:
        return '\n'.join(map(render_pretty_expression, exprs))
    res = map(render_expression, exprs)
    if options.args().wrap_lines:
        res = [line for r in res for line in textwrap.wrap(r, width = 78, subsequent_indent = '  ')]
    return '\n'.join(res)

def write_smtlib_to_file(exprs, filename):
    """Writes a sequence of nodes to a file."""
    open(filename, 'w').write(render_smtlib(exprs))
